@@ -8,31 +8,44 @@
 *   03/10/05   DGP   Change to preserve load order of object.
 *   03/15/05   DGP   Add back ctlbits, oops.
 *   03/16/05   DGP   Added Segment option.
+*   12/15/05   RMS   MINGW changes.
+*   03/19/07   DGP   Include parity tables directly.
+*   03/31/10   DGP   Correct DECREMENT relocation.
+*   10/20/10   DGP   Added -b, NO BOM, option.
+*   01/03/12   DGP   Fixed ctlbits at start of block.
 *	
 ***********************************************************************/
 
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <memory.h>
 #include <errno.h>
+#if !defined(WIN32)
+#include <unistd.h>
+#endif
 
 #include "sysdef.h"
 #include "objdef.h"
-#include "parity.h"
 
-#ifdef WIN32
+#if defined(WIN32) && !defined(MINGW)
 #define IOCP 0400000000000I64
 #define IOCT 0500000000000I64
+#define TRA  0002000000000I64
 #else
 #define IOCP 0400000000000ULL
 #define IOCT 0500000000000ULL
+#define TRA  0002000000000ULL
 #endif
 
 static t_uint64 memory[MEMSIZE];
 
+static int nobom;
 static uint8 word[12];
 static uint8 ctlbits;
 
+#include "parity.h"
+
 /***********************************************************************
 * writeword - write a word.
 ***********************************************************************/
@@ -70,9 +83,12 @@ writeblock (int loadaddr, int wrdcnt, FILE *fd)
    ** Write out loader control word
    */
 
-   wrd1 = IOCP | ((t_uint64)wrdcnt << DECRSHIFT) |
-	 (t_uint64)loadaddr;
-   writeword (wrd1, fd);
+   if (!nobom)
+   {
+      wrd1 = IOCP | ((t_uint64)wrdcnt << DECRSHIFT) |
+	    (t_uint64)loadaddr;
+      writeword (wrd1, fd);
+   }
 
    /*
    ** Write the corresponding block
@@ -89,6 +105,7 @@ writeblock (int loadaddr, int wrdcnt, FILE *fd)
 * Main procedure
 ***********************************************************************/
 
+int
 main (int argc, char *argv[])
 {
    FILE *infd;
@@ -104,6 +121,7 @@ main (int argc, char *argv[])
    int wrdcnt;
    int noeom;
    int overlay;
+   int verbose;
    int i;
    char inbuf[82];
 
@@ -112,7 +130,9 @@ main (int argc, char *argv[])
    */
 
    noeom = FALSE;
+   nobom = FALSE;
    overlay = FALSE;
+   verbose = FALSE;
 
    infile = NULL;
    outfile = NULL;
@@ -134,13 +154,17 @@ main (int argc, char *argv[])
       {
          for (bp++; *bp; bp++) switch (*bp)
 	 {
+	 case 'b': /* No BOM */
+	    nobom = TRUE;
+	    break;
+
 	 case 'e': /* No EOM */
 	    noeom = TRUE;
 	    break;
 
 	 case 'l': /* Load address, for relocatable images */
 	    i++;
-	    curraddr = loadaddr = atoi(argv[i]);
+	    curraddr = loadaddr = atoi (argv[i]);
 	    break;
 
 	 case 'o': /* Overlay */
@@ -149,7 +173,11 @@ main (int argc, char *argv[])
 
 	 case 's': /* Object segment number */
 	    i++;
-	    objindex = atoi(argv[i]);
+	    objindex = atoi (argv[i]);
+	    break;
+
+	 case 'v': /* Verbose */
+	    verbose = TRUE;
 	    break;
 
 	 default:
@@ -165,6 +193,7 @@ main (int argc, char *argv[])
       usage:
 	 fprintf (stderr, "usage: obj2bin [options] infile outfile\n");
 	 fprintf (stderr, " options:\n");
+	 fprintf (stderr, "    -b           - No BOM word\n");
 	 fprintf (stderr, "    -e           - No EOM word\n");
 	 fprintf (stderr, "    -l addr      - Load address\n");
 	 fprintf (stderr, "    -o           - Add as overlay\n");
@@ -181,17 +210,17 @@ main (int argc, char *argv[])
 
    if ((infd = fopen (infile, "r")) == NULL)
    {
-      fprintf (stderr, "obj2bin: input open failed: %s",
-	       strerror(errno));
-      fprintf (stderr, "filename: %s", infile);
+      fprintf (stderr, "obj2bin: input open failed: %s\n",
+	       strerror (errno));
+      fprintf (stderr, "filename: %s\n", infile);
       exit (1);
    }
 
    if ((outfd = fopen (outfile, overlay ? "ab" : "wb")) == NULL)
    {
-      fprintf (stderr, "obj2bin: output open failed: %s",
-	       strerror(errno));
-      fprintf (stderr, "filename: %s", outfile);
+      fprintf (stderr, "obj2bin: output open failed: %s\n",
+	       strerror (errno));
+      fprintf (stderr, "filename: %s\n", outfile);
       exit (1);
    }
 
@@ -201,7 +230,7 @@ main (int argc, char *argv[])
 
    for (i = 0; i < objindex; i++)
    {
-      while ((bp = fgets (inbuf, sizeof(inbuf), infd)) != NULL)
+      while ((bp = fgets (inbuf, sizeof (inbuf), infd)) != NULL)
       {
          if (*bp == IBSYSSYM) break;
       }
@@ -220,10 +249,9 @@ main (int argc, char *argv[])
 
    memset (&memory, '\0', sizeof (memory));
 
-   while (fgets (inbuf, sizeof(inbuf), infd))
+   while (fgets (inbuf, sizeof (inbuf), infd))
    {
       char *op = inbuf;
-      int relo;
 
       if (*op == IBSYSSYM) break; /* $EOF */
 
@@ -231,15 +259,15 @@ main (int argc, char *argv[])
       {
 	 char otag;
 	 char item[16];
-	 t_uint64 lldata;
+	 t_uint64 lldata, dectmp;
 
 	 otag = *op++;
 	 if (otag == ' ') break;
 	 lldata = 0;
 	 strncpy (item, op, 12);
 	 item[12] = '\0';
-#ifdef WIN32
-	 sscanf (item, "%i64o", &lldata);
+#if defined(WIN32)
+	 sscanf (item, "%I64o", &lldata);
 #else
 	 sscanf (item, "%llo", &lldata);
 #endif
@@ -248,8 +276,8 @@ main (int argc, char *argv[])
 	 printf ("loadaddr = %05o, curraddr = %05o\n",
 		  loadaddr, curraddr);
 	 printf ("   otag = %c, item = %s\n", otag, item);
-#ifdef WIN32
-	 printf ("   lldata = %12.12i64o\n", lldata);
+#if defined(WIN32)
+	 printf ("   lldata = %12.12I64o\n", lldata);
 #else
 	 printf ("   lldata = %12.12llo\n", lldata);
 #endif
@@ -265,6 +293,10 @@ main (int argc, char *argv[])
 	    }
 	    curraddr = loadaddr = (int)(lldata & ADDRMASK);
 	    startaddr = curraddr;
+	    if (verbose)
+	    {
+	       printf ("obj2bin: ABSORG = %05o\n", curraddr);
+	    }
 	    break;
 
 	 case RELORG_TAG:
@@ -273,8 +305,12 @@ main (int argc, char *argv[])
 	       writeblock (startaddr, wrdcnt, outfd);
 	       wrdcnt = 0;
 	    }
-	    curraddr = (int)(lldata + loadaddr) & ADDRMASK;
+	    curraddr = (int)((lldata + loadaddr) & ADDRMASK);
 	    startaddr = curraddr;
+	    if (verbose)
+	    {
+	       printf ("obj2bin: RELORG = %05o\n", curraddr);
+	    }
 	    break;
 
 	 case BSS_TAG:
@@ -286,6 +322,10 @@ main (int argc, char *argv[])
 	    curraddr += (int)(lldata & ADDRMASK);
 	    curraddr &= ADDRMASK;
 	    startaddr = curraddr;
+	    if (verbose)
+	    {
+	       printf ("obj2bin: BSS = %05o\n", curraddr);
+	    }
 	    break;
 
 	 case ABSDATA_TAG:
@@ -303,9 +343,9 @@ main (int argc, char *argv[])
 	 case RELBOTH_TAG:
 	    lldata += loadaddr;
 	 case RELDECR_TAG:
-	    relo = ((lldata & DECRMASK) >> DECRSHIFT) + loadaddr;
+	    dectmp = ((lldata & DECRMASK) >> DECRSHIFT) + loadaddr;
 	    lldata &= ~DECRMASK;
-	    lldata |= (relo << DECRSHIFT); 
+	    lldata |= (dectmp << DECRSHIFT); 
 	    memory[curraddr] = lldata;
 	    curraddr++;
 	    wrdcnt++;
@@ -315,12 +355,20 @@ main (int argc, char *argv[])
 	    goto LOADED;
 	 case ABSENTRY_TAG:
 	    entaddr = (int)(lldata & ADDRMASK);
+	    if (verbose)
+	    {
+	       printf ("obj2bin: ABSENTRY = %05o\n", entaddr);
+	    }
 	    break;
 
 	 case RELXFER_TAG:
 	    goto LOADED;
 	 case RELENTRY_TAG:
 	    entaddr = (int)(lldata & ADDRMASK) + loadaddr;
+	    if (verbose)
+	    {
+	       printf ("obj2bin: RELENTRY = %05o\n", entaddr);
+	    }
 	    break;
 
 	 default: ;
@@ -341,7 +389,9 @@ LOADED:
    */
 
    if (!noeom)
-      writeword (IOCT, outfd);
+   {
+      writeword (IOCT | (entaddr >= 0 ? entaddr : 0), outfd);
+   }
 
    fclose (infd);
    fclose (outfd);
